from skimage import io, color
from PIL import Image

import numpy as np
import cv2

from scipy.optimize import minimize
from scipy.optimize import Bounds

try:
    from utils import *
except ImportError:
    from .utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)


def convert_constraints_and_weights_data( pixel_lum_constraints, curve_lum_constraints, cons_weights, num_p, N=100 ):
    '''
    Given:
        `pixel_lum_constraints`: A #c-by-2 array of luminance constraints
        `curve_lum_constraints`: Like: [ (palette 0, 0.4, 0.2), (palette 1, 0.3, 0.5), ... ]
        `cons_weights`: A #p-by-#c array of weights at the constraints
    '''
    pl_cons, cl_cons = [], []
    
    # convert luminance constraints from pixels
    for c in pixel_lum_constraints:
        pl_cons.append( (int(c[0]*N), c[1]) )
    
    # convert luminance constraints from curves themselves
    for c in curve_lum_constraints:
        cl_cons.append( (c[0], int(c[1]*N), c[2]) )
    
    if cons_weights.shape[0] != 0:
        cons_weights[-1, :] /= 100
        cons_weights = cons_weights / np.sum(cons_weights, axis=0)
    else:   # if currently no pixel weigh constraints, just make it #p-by-1 zero matrix
        cons_weights = np.zeros( (num_p, 1) ) 
    return pl_cons, cl_cons, cons_weights


def sparse_optimization( laplacian, constraints, curve_constraints, W, N, w_c ):
    '''
    Given:
        `laplacian`: A N-by-N laplace operator with mirroring boundary
        `constraints`: A list of luminance constraints, i.e. [ (int, 0.2), (int, 0.3), ... ]
        `curve_constraints`: A list of curve constraints, i.e. [ (int, 0.2), (int, 0.3), ... ]
        `W`: A #p-by-#c array of weights at constraints
        `N`: Number of samples (100 is enough)
        `w_c`: A scalar of the penalization term
    
    Return:
        `opt_lum`: A N-by-#p array of optimized luminance at sampled points
    '''
    np.set_printoptions( suppress = True )

    num_p = W.shape[0]
    num_cons = W.shape[1]
    
    print( 'Luminance constraints: ', constraints )
    print( 'Curve constraints: ', curve_constraints )
    
    # define selection and constraint matrices for image-space constraints
    dim_I = len( constraints ) if len( constraints ) != 0 else 1
    C_I = np.zeros( ( N, dim_I ) )
    S_I = np.zeros( ( N, dim_I ) )
    for i, ( Lindex, val ) in enumerate( constraints ):
        C_I[Lindex, i] = val
        S_I[Lindex, i] = 1
    
    # define selection and constraint matrices for constraints for direct curve manipulation
    C_d = np.zeros( ( N, num_p ) )
    S_d = np.zeros( ( N, num_p ) )
    for c in curve_constraints:
        C_d[ c[1], c[0] ] = c[2]
        S_d[ c[1], c[0] ] = 1
    
    def gradients( para ):
        '''
        Given:
            `para`: A (#p)*(N) vector. Each (#p) portion of the first (#p)*N elements in the vector
                    is luminance at sampled points.
        Return:
            `grad`: gradient with respect to the given parameters
        '''
        luminance = np.reshape( para[: num_p*N], (N, num_p), order = 'F' )
        grad_lum = np.zeros( (N, num_p) )
        
        # gradient of sparsity term for luminances and palettes
        for i in range( num_p ):
            BtBL = laplacian.T @ laplacian @ luminance[:, i]
            sparse = np.sqrt( luminance[:, i].T @ BtBL )
            
            lum_sparsity_term = BtBL / sparse
            grad_lum[:, i] += lum_sparsity_term
            
        # gradient of constraint term for luminances and palettes
        for i in range( num_p ):
            for j in range( num_cons ):
                recon_lum = 0
                for k in range( num_p ):
                    recon_lum += W[k, j] * luminance[:, k]
                diff_lum = np.diag( S_I[:, j] ) @ ( np.diag( S_I[:, j] ) @ recon_lum - C_I[:, j] )
                grad_lum[:, i] += 2 * w_c* W[i, j] * diff_lum
            grad_lum[:, i] += 2 * w_c* np.diag( S_d[:, i] ) @ ( np.diag( S_d[:, i] ) @ luminance[:, i] - C_d[:, i] )
        
        grad_lum = grad_lum.flatten( 'F' )
        return grad_lum/100
        
    def func( para ):
        '''
        Given:
            `para`: A (#p)*(N+2) vector. Each (#p) portion of the first (#p)*N elements in the vector
                    is luminance at sampled points. The rest 2*(#p) is delta-palette
        Return:
            `loss`: A loss measuring the sparsity together of both luminances and palette changes
        '''
        loss_sp, loss_lum = 0, 0
        luminance = np.reshape( para[: num_p*N], (N, num_p), order = 'F' )
        
        # sparsity loss
        bilap = laplacian.T @ laplacian
        for i in range( num_p ):
            loss_sp += np.sqrt( luminance[:, i].T @ bilap @ luminance[:, i] )
        
        # luminance and color constraint loss
        ## First term
        for j in range( num_cons ):
            recon_lum = 0
            for i in range( num_p ):
                recon_lum += W[i, j]*luminance[:, i]
            loss_lum += np.linalg.norm( np.multiply( S_I[:, j], recon_lum ) - C_I[:, j] ) ** 2
        
        ## Second term
        for i in range( num_p ):
            loss_lum += np.linalg.norm( np.multiply( S_d[:, i], luminance[:, i] ) - C_d[:, i] ) ** 2
        
        return (loss_sp + w_c*loss_lum)/100
    
    # constraint settings
    def lum_boundary_0( para ):
        luminance = np.reshape( para[: num_p*N], (N, num_p), order = 'F' )
        return 0.01 - np.linalg.norm( luminance[0, :] )
    
    def lum_boundary_1( para ):
        luminance = np.reshape( para[: num_p*N], (N, num_p), order = 'F' )
        return 0.01 - np.linalg.norm( luminance[-1, :] - 1 )
    
    cons = [ {'type': 'ineq', 'fun': lum_boundary_0},
             {'type': 'ineq', 'fun': lum_boundary_1} ]
    
    # initial guess
    x0 = np.zeros( N*num_p )
    for i in range( num_p ):
        x0[ (i+1) * N - 1 ] = 1
    
    res = minimize( func, x0, method='SLSQP', jac=gradients, constraints=cons, options={'ftol': 1e-5, 'disp': False, 'maxiter': 2000} )
    para = res.x
    opt_luminance = np.reshape( para[: num_p*N], (N, num_p), order = 'F' )
    
    print( res )
    return opt_luminance
    

def get_laplacian( N ):
    '''
    Given:
        `N`: number of samples

    Return:
        `Lap`: Laplace operator
    '''
    def grad( n ):
        G = np.zeros( ( n-1, n ) )
        for i in range( 0, n-1 ):
            G[i,i+1] = 1
            G[i,i] = -1
        return G
    
    # building bilaplacian
    G = grad( N )
    Lap = G.T @ G
    
    # mirroring the laplacian to avoid prefering 0 gradients at endpoints
    Lap[0,:] = 0
    Lap[-1,:] = 0
    
    return Lap


def compute_new_luminance( L0, L_bilap, weights, N=100 ):
    '''
    Given:
        `L0`: A N-by-M array of original luminance
        `L_bilap`: A N-by-#P array of bilaplacian solutions
        `weights`: A N-by-#P array of weights
    
    Return:
        Reconstructed Luminance
    '''
    
    from scipy import interpolate
    
    # ignore weights for grey and re-normalize it
    lum_weights = np.copy( weights )
    lum_weights[:, -1] /= 100
    lum_weights = (lum_weights.T / np.sum(lum_weights, axis=1)).T
    
    L_new = 0
    L = np.linspace( 0, 1, N )
    L0_ = L0.reshape( -1 )
    for i in range( lum_weights.shape[1] ):
        f = interpolate.interp1d( L, L_bilap[:,i] )
        L_new += np.multiply( f( L0_ ), lum_weights[:, i] )
    return L_new.reshape( L0.shape )

def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'optimization on entire sparsity.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    L0 = img[:, :, 0] / 100
    ab = img[:, :, 1:]
    
    ab_palette = extract_AB_palette( img, 4 )
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], 5) )
    
    N = 10 # number of samples
    
    i, j = 400, 700
    print( img[i,j] )
    target_lum = 0.802
    lum_cons = np.array( [[L0[i, j], target_lum]] )
    p_weights = np.array( [ weights[i, j] ] )
    #curve_cons = [ (0, 0.4, 0.2) ]
    curve_cons = []
    pixel_lum_cons, curve_lum_cons, cons_weights = convert_constraints_and_weights_data( lum_cons, curve_cons, p_weights.T, ab_palette.shape[0], N )
    
    laplacian = get_laplacian( N )
    w_c = 100
    
    import time
    start_time = time.time()
    opt_lum = sparse_optimization( laplacian, pixel_lum_cons, curve_lum_cons, cons_weights, N, w_c )
    print( 'Runtime: ', time.time() - start_time )
    
    recon_ab = weights @ ab_palette
    recon_lum = compute_new_luminance( L0, opt_lum, weights.reshape(-1, ab_palette.shape[0]), N ) * 100
    
    recon_img = np.zeros_like( img )
    recon_img[:, :, 0] = recon_lum
    recon_img[:, :, 1:] = recon_ab
    
    print( 'Reconstruction error: ', np.abs( recon_img[i,j][0] - target_lum*100 ) )
    recon_img = ( color.lab2rgb( recon_img ) * 255. ).clip( 0, 255 ).astype( np.uint8 )
    Image.fromarray( recon_img ).save( '../test-edit-sqrt-' + str(N) + '.png' )
    
            
if __name__ == '__main__':
    main()
    