from skimage import io, color
from PIL import Image

import numpy as np
import cv2

from scipy.interpolate import pchip_interpolate
from scipy.optimize import minimize
from scipy.optimize import Bounds

try:
    from utils import *
except ImportError:
    from .utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def color_optimization( palette, weights, target_color, palette_cons, verbose=False ):
    '''
    Given:
        `palette`: A #p-by-2 array
        `weights`: A array of shape (C, #p)
        `target_colors`: A array of shape (C, 2)
        `palette_cons`: A list of palette constraints. Like: [ [j1, (a1,b1)], [j2, (a2,b2)], ... ]
        `verbose`: toggle for displaying optimization results
    
    Return:
        `palette_opt`: A #p-by-2 array from optimization results
    '''
    
    if np.size( weights ) != 0:
        assert palette.shape[0] == weights.shape[1]
        assert palette.shape[1] == target_color.shape[1]
        assert weights.shape[0] == target_color.shape[0]
    else:
        weights = np.zeros( (1,palette.shape[0]) )
        target_color = np.zeros( (1,2) )
        
    # set up palette constraints
    pc = np.zeros( (len(palette_cons), 2) ) 
    ind = np.zeros( len(palette_cons) ).astype( int )
    for i, cons in palette_cons:
        pc[i, :] = cons
        ind[i] = i
    
    ### objective function
    def func( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        return np.sum( np.linalg.norm( delp, axis = 1 ) )     # optimize for palette's sparsity
    
    ### image-space constraint
    def color_constraints( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        rec_colors = weights @ ( palette + delp )
        return np.linalg.norm( rec_colors - target_color, axis = 1 )
    
    ### constraints on palette color
    def palette_constraints( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        new_palette = palette + delp
        return np.linalg.norm( new_palette[ ind ] - pc, axis = 1 )
    
    bounds = Bounds( -128 - palette.flatten(), 127 - palette.flatten() )
    
    if len( palette_cons ) == 0:
        cons = [ {'type': 'eq', 'fun': color_constraints} ]
    else:
        # use inequality on color constraints to avoid singularity
        # (in practice/numerically, it's ok since our sparsity energy is going to dominate this.)
        cons = [ {'type': 'ineq', 'fun': color_constraints}, {'type': 'eq', 'fun': palette_constraints} ]
    
    x0 = np.zeros( palette.shape[0] * palette.shape[1] )
    res = minimize( func, x0, method='SLSQP', constraints=cons, options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    delp = ( res.x ).reshape( palette.shape )
    
    if verbose:
        print( '-- Original weights:\n', weights )
        print( '-- Original color:\n', weights @ palette )
        print( '-- Target color:\n', target_color )
        print( '-- Reconstructed color:\n', weights @ ( delp + palette ) )
        
    return delp + palette

def lum_optimization( L0, L_target, control_palette, weights, verbose=False ):
    '''
    Given:
        `L0`: A C-by-1 array of original luminances (range: [0, 1])
        `L_target`: A C-by-1 array of target luminances corresponding to L0 (range: [0, 1])
        `control_palette`: A #P-by-C-2 array of controls points for each palette
        `weights`: A array of shape (C, #p)
        `verbose`: toggle for displaying optimization results
    
    Return:
        `new_control_palette`: A #P-by-C-2 array of new `control_palette`
    '''
    np.set_printoptions( suppress = True )
    assert len( control_palette.shape ) == 3
    assert L0.shape[0] == L_target.shape[0] == weights.shape[0]
    
    num_cons = L0.shape[0]
    num_func = control_palette.shape[0]
    num_fixed = control_palette.shape[1]
    
    # initial function values at L0 for each shape function
    discrete_vals = np.zeros( (num_cons, num_func) )
    for c in range( num_cons ):
        for i in range( num_func ):
            discrete_vals[c][i] = pchip_interpolate( control_palette[i][:,0], control_palette[i][:,1], L0[c] )
    
    # initialization for vertical changes for each shape function
    delta_function = np.ones_like( discrete_vals ).flatten()
    
    # ignore weights for grey and re-normalize it
    lum_weights = np.copy( weights )
    lum_weights[:, -1] /= 1000
    lum_weights = (lum_weights.T / np.sum(lum_weights, axis=1)).T
    
    def loss( delta_func ):
        # L2,1-norm
        delta_func = delta_func.reshape( discrete_vals.shape )
        return np.sum( np.linalg.norm( delta_func, axis = 0 ) )
    
    def lum_constraints( delta_func ):
        # Σ_i w_i (f_i + Δf_i)(L0) = L'
        delta_func = delta_func.reshape( discrete_vals.shape )
        return np.sum( lum_weights * ( discrete_vals + delta_func ), axis = 1 ) - L_target
    
    # -f_i(L0) <= Δf_i(L0) <= 1 - f_i(L0)
    bounds = Bounds( -discrete_vals.flatten(), 1-discrete_vals.flatten() )
    cons = [ {'type': 'eq', 'fun': lum_constraints} ]
    
    res = minimize( loss, delta_function, method='SLSQP', constraints=cons, options={'ftol': 1e-8, 'disp': False, 'maxiter': 300}, bounds=bounds )
    delta_function = ( res.x ).reshape( discrete_vals.shape )
    
    if verbose:
        print( '-- Original weights:\n', weights )
        print( '-- Renormalized weights:\n', lum_weights )
        print( '-- Original luminance:\n', L0 )
        print( '-- Target luminance:\n', L_target )
        print( '-- Reconstructed error:\n', lum_constraints( delta_function ) )
    
    # add optimized offsets to existing constraints
    ind = list( np.arange(L0.shape[0]) )
    for i in range( L0.shape[0] ):
        if L0[i] in control_palette[0][:, 0]:
            ind.pop( ind.index( i ) )
            
    # update control points palette
    new_control_pts_x = np.tile( L0[ ind ], num_func )
    new_control_pts_y = np.ravel( delta_function[ ind, : ] + discrete_vals[ ind, : ], order = 'F' )
    new_control_pts = np.vstack( (new_control_pts_x, new_control_pts_y) ).T.reshape( (num_func, len( ind ), 2) )
    
    # create new control palette
    new_control_palette = np.concatenate( (control_palette, new_control_pts), axis=1 )
    return new_control_palette

def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'shadow/highlight testing.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    ab_palette = extract_AB_palette( img, 4 )   # palette size here
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], ab_palette.shape[0]) )
    
    L0 = img[:, :, 0] / 100
    ab = weights @ ab_palette
    
    i, j = 100, 52 
    p_weights = np.array( [weights[i, j]] )
    
    target_color = np.array( [ab[i, j]] )
    target_lum = 0.6
    
    print( 'Original luminance: ', L0[i, j] )
    print( 'Target luminance: ', target_lum )
    
    print( 'Original color: ', ab[i, j] )
    print( 'Target color: ', target_color )
    
    ###-------
    
    # optimize color palette
    palette_target = [ [0, (0, 20)] ]
    palette_opt = color_optimization( ab_palette, p_weights, target_color, palette_target )
    
    # tiles up fixed points for each shape function ( we need to pin down (0,0) and (1,1) )
    control_palette = np.tile( np.array( [[0,0], [1,1]] ), (5, 1, 1) )
    
    L_constraints = np.array( [[L0[i,j], target_lum]] ) # An C-by-2 array of constraints
    
    # hard-code testing
    p_weights = np.array([[ 0.54955942,  0,  0.09129066, 0.        ,  0.35914992]])
    L_constraints = np.array( [[0.49097798, 0.42968466]] )
    
    control_palette = lum_optimization( L_constraints[:, 0], L_constraints[:, 1], control_palette, p_weights, True )
    
    # reconstruct back luminance
    recon_lum = get_recon_lum( control_palette, weights.reshape(-1,5), L0 ) * 100
    recon_img = get_recon_img( palette_opt, weights, recon_lum )
    print( 'Final reconstructed lum: ', recon_lum[i,j] / 100 )
    
    ### save result
    recon_img = ( color.lab2rgb( recon_img ) * 255. ).clip( 0, 255 ).astype( np.uint8 )
    Image.fromarray( recon_img ).save( '../test-edit.png' )
    
if __name__ == '__main__':
    main()